package BoardStreams;
use 5.008001;
use strict;
use warnings;

our $VERSION = "v0.0.3";



1;
__END__

=encoding utf-8

=head1 NAME

BoardStreams - It's new $module

=head1 SYNOPSIS

    use BoardStreams;

=head1 DESCRIPTION

BoardStreams is ...

=head1 LICENSE

Copyright (C) Alexander Karelas.

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.

=head1 AUTHOR

Alexander Karelas E<lt>karjala@cpan.orgE<gt>

=cut

